package sun.nio.ch;

import java.net.NetworkInterface;
import java.nio.channels.DatagramChannel;
import java.net.StandardSocketOptions;
import java.net.SocketOption;
import java.nio.channels.ClosedChannelException;
import java.net.SocketTimeoutException;
import java.nio.channels.IllegalBlockingModeException;
import java.nio.ByteBuffer;
import java.net.DatagramPacket;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.SocketException;
import java.net.SocketAddress;
import java.io.IOException;
import java.net.DatagramSocketImpl;
import java.net.DatagramSocket;

public class DatagramSocketAdaptor extends DatagramSocket
{
    private final DatagramChannelImpl dc;
    private volatile int timeout;
    private static final DatagramSocketImpl dummyDatagramSocket;
    
    private DatagramSocketAdaptor(DatagramChannelImpl dc) throws IOException {
        super(DatagramSocketAdaptor.dummyDatagramSocket);
        this.timeout = 0;
        this.dc = dc;
    }
    
    public static DatagramSocket create(DatagramChannelImpl datagramChannelImpl) {
        try {
            return new DatagramSocketAdaptor(datagramChannelImpl);
        }
        catch (IOException cause) {
            throw new Error(cause);
        }
    }
    
    private void connectInternal(SocketAddress socketAddress) throws SocketException {
        int port = Net.asInetSocketAddress(socketAddress).getPort();
        if (port < 0 || port > 65535) {
            throw new IllegalArgumentException("connect: " + port);
        }
        if (socketAddress == null) {
            throw new IllegalArgumentException("connect: null address");
        }
        if (this.isClosed()) {
            return;
        }
        try {
            this.dc.connect(socketAddress);
        }
        catch (Exception ex) {
            Net.translateToSocketException(ex);
        }
        return;
    }
    
    @Override
    public void bind(SocketAddress socketAddress) throws SocketException {
        try {
            if (socketAddress == null) {
                socketAddress = new InetSocketAddress(0);
            }
            this.dc.bind(socketAddress);
        }
        catch (Exception ex) {
            Net.translateToSocketException(ex);
        }
    }
    
    @Override
    public void connect(InetAddress addr, int port) {
        try {
            this.connectInternal(new InetSocketAddress(addr, port));
        }
        catch (SocketException ex) {}
    }
    
    @Override
    public void connect(SocketAddress socketAddress) throws SocketException {
        if (socketAddress == null) {
            throw new IllegalArgumentException("Address can't be null");
        }
        this.connectInternal(socketAddress);
        return;
    }
    
    @Override
    public void disconnect() {
        try {
            this.dc.disconnect();
        }
        catch (IOException cause) {
            throw new Error(cause);
        }
    }
    
    @Override
    public boolean isBound() {
        return this.dc.localAddress() != null;
    }
    
    @Override
    public boolean isConnected() {
        return this.dc.remoteAddress() != null;
    }
    
    @Override
    public InetAddress getInetAddress() {
        return this.isConnected() ? Net.asInetSocketAddress(this.dc.remoteAddress()).getAddress() : null;
    }
    
    @Override
    public int getPort() {
        return this.isConnected() ? Net.asInetSocketAddress(this.dc.remoteAddress()).getPort() : -1;
    }
    
    @Override
    public SocketAddress getLocalSocketAddress() {
        InetSocketAddress inetSocketAddress = (InetSocketAddress)this.dc.localAddress();
        if (inetSocketAddress == null || this.isClosed()) {
            return null;
        }
        InetAddress address = inetSocketAddress.getAddress();
        if (address.isAnyLocalAddress()) {
            return inetSocketAddress;
        }
        SecurityManager securityManager = System.getSecurityManager();
        if (securityManager != null) {
            try {
                securityManager.checkConnect(address.getHostAddress(), -1);
            }
            catch (SecurityException ex) {
                return new InetSocketAddress(inetSocketAddress.getPort());
            }
        }
        return inetSocketAddress;
    }
    
    @Override
    public void send(DatagramPacket datagramPacket) throws IOException {
        synchronized (this.dc.blockingLock()) {
            if (!this.dc.isBlocking()) {
                throw new IllegalBlockingModeException();
            }
            try {
                synchronized (datagramPacket) {
                    ByteBuffer wrap = ByteBuffer.wrap(datagramPacket.getData(), datagramPacket.getOffset(), datagramPacket.getLength());
                    if (this.dc.isConnected()) {
                        if (datagramPacket.getAddress() == null) {
                            InetSocketAddress inetSocketAddress = (InetSocketAddress)this.dc.remoteAddress();
                            datagramPacket.setPort(inetSocketAddress.getPort());
                            datagramPacket.setAddress(inetSocketAddress.getAddress());
                            this.dc.write(wrap);
                        }
                        else {
                            this.dc.send(wrap, datagramPacket.getSocketAddress());
                        }
                    }
                    else {
                        this.dc.send(wrap, datagramPacket.getSocketAddress());
                    }
                }
            }
            catch (IOException ex) {
                Net.translateException(ex);
            }
        }
    }
    
    private SocketAddress receive(ByteBuffer byteBuffer) throws IOException {
        if (this.timeout == 0) {
            return this.dc.receive(byteBuffer);
        }
        this.dc.configureBlocking(false);
        try {
            SocketAddress receive;
            if ((receive = this.dc.receive(byteBuffer)) != null) {
            }
            else {
                long n = (long)this.timeout;
                while (this.dc.isOpen()) {
                    long currentTimeMillis = System.currentTimeMillis();
                    int poll = this.dc.poll((int)Net.POLLIN, n);
                    SocketAddress receive2;
                    if (poll > 0 && (poll & Net.POLLIN) != 0x0 && (receive2 = this.dc.receive(byteBuffer)) != null) {
                    }
                    else {
                        n -= System.currentTimeMillis() - currentTimeMillis;
                        if (n <= 0L) {
                            throw new SocketTimeoutException();
                        }
                        continue;
                    }
                }
                throw new ClosedChannelException();
            }
        }
        finally {
            if (this.dc.isOpen()) {
                this.dc.configureBlocking(true);
            }
        }
    }
    
    @Override
    public void receive(DatagramPacket datagramPacket) throws IOException {
        synchronized (this.dc.blockingLock()) {
            if (!this.dc.isBlocking()) {
                throw new IllegalBlockingModeException();
            }
            try {
                synchronized (datagramPacket) {
                    ByteBuffer wrap = ByteBuffer.wrap(datagramPacket.getData(), datagramPacket.getOffset(), datagramPacket.getLength());
                    datagramPacket.setSocketAddress(this.receive(wrap));
                    datagramPacket.setLength(wrap.position() - datagramPacket.getOffset());
                }
            }
            catch (IOException ex) {
                Net.translateException(ex);
            }
        }
    }
    
    @Override
    public InetAddress getLocalAddress() {
        if (this.isClosed()) {
            return null;
        }
        SocketAddress localAddress = this.dc.localAddress();
        if (localAddress == null) {
            localAddress = new InetSocketAddress(0);
        }
        InetAddress address = ((InetSocketAddress)localAddress).getAddress();
        SecurityManager securityManager = System.getSecurityManager();
        if (securityManager != null) {
            try {
                securityManager.checkConnect(address.getHostAddress(), -1);
            }
            catch (SecurityException ex) {
                return new InetSocketAddress(0).getAddress();
            }
        }
        return address;
    }
    
    @Override
    public int getLocalPort() {
        if (this.isClosed()) {
            return -1;
        }
        try {
            SocketAddress localAddress = this.dc.getLocalAddress();
            if (localAddress != null) {
                return ((InetSocketAddress)localAddress).getPort();
            }
        }
        catch (Exception ex) {}
        return 0;
    }
    
    @Override
    public void setSoTimeout(int timeout) throws SocketException {
        this.timeout = timeout;
    }
    
    @Override
    public int getSoTimeout() throws SocketException {
        return this.timeout;
    }
    
    private void setBooleanOption(SocketOption<Boolean> socketOption, boolean b) throws SocketException {
        try {
            this.dc.setOption(socketOption, Boolean.valueOf(b));
        }
        catch (IOException ex) {
            Net.translateToSocketException(ex);
        }
    }
    
    private void setIntOption(SocketOption<Integer> socketOption, int i) throws SocketException {
        try {
            this.dc.setOption(socketOption, Integer.valueOf(i));
        }
        catch (IOException ex) {
            Net.translateToSocketException(ex);
        }
    }
    
    private boolean getBooleanOption(SocketOption<Boolean> socketOption) throws SocketException {
        try {
            return (boolean)this.dc.getOption(socketOption);
        }
        catch (IOException ex) {
            Net.translateToSocketException(ex);
            return false;
        }
    }
    
    private int getIntOption(SocketOption<Integer> socketOption) throws SocketException {
        try {
            return (int)this.dc.getOption(socketOption);
        }
        catch (IOException ex) {
            Net.translateToSocketException(ex);
            return -1;
        }
    }
    
    @Override
    public void setSendBufferSize(int n) throws SocketException {
        if (n <= 0) {
            throw new IllegalArgumentException("Invalid send size");
        }
        this.setIntOption(StandardSocketOptions.SO_SNDBUF, n);
        return;
    }
    
    @Override
    public int getSendBufferSize() throws SocketException {
        return this.getIntOption(StandardSocketOptions.SO_SNDBUF);
    }
    
    @Override
    public void setReceiveBufferSize(int n) throws SocketException {
        if (n <= 0) {
            throw new IllegalArgumentException("Invalid receive size");
        }
        this.setIntOption(StandardSocketOptions.SO_RCVBUF, n);
        return;
    }
    
    @Override
    public int getReceiveBufferSize() throws SocketException {
        return this.getIntOption(StandardSocketOptions.SO_RCVBUF);
    }
    
    @Override
    public void setReuseAddress(boolean b) throws SocketException {
        this.setBooleanOption(StandardSocketOptions.SO_REUSEADDR, b);
    }
    
    @Override
    public boolean getReuseAddress() throws SocketException {
        return this.getBooleanOption(StandardSocketOptions.SO_REUSEADDR);
    }
    
    @Override
    public void setBroadcast(boolean b) throws SocketException {
        this.setBooleanOption(StandardSocketOptions.SO_BROADCAST, b);
    }
    
    @Override
    public boolean getBroadcast() throws SocketException {
        return this.getBooleanOption(StandardSocketOptions.SO_BROADCAST);
    }
    
    @Override
    public void setTrafficClass(int n) throws SocketException {
        this.setIntOption(StandardSocketOptions.IP_TOS, n);
    }
    
    @Override
    public int getTrafficClass() throws SocketException {
        return this.getIntOption(StandardSocketOptions.IP_TOS);
    }
    
    @Override
    public void close() {
        try {
            this.dc.close();
        }
        catch (IOException cause) {
            throw new Error(cause);
        }
    }
    
    @Override
    public boolean isClosed() {
        return !this.dc.isOpen();
    }
    
    @Override
    public DatagramChannel getChannel() {
        return this.dc;
    }
    
    static {
        dummyDatagramSocket = new DatagramSocketImpl() {
            @Override
            protected void create() throws SocketException {
            }
            
            @Override
            protected void bind(int n, InetAddress inetAddress) throws SocketException {
            }
            
            @Override
            protected void send(DatagramPacket datagramPacket) throws IOException {
            }
            
            @Override
            protected int peek(InetAddress inetAddress) throws IOException {
                return 0;
            }
            
            @Override
            protected int peekData(DatagramPacket datagramPacket) throws IOException {
                return 0;
            }
            
            @Override
            protected void receive(DatagramPacket datagramPacket) throws IOException {
            }
            
            @Deprecated
            @Override
            protected void setTTL(byte b) throws IOException {
            }
            
            @Deprecated
            @Override
            protected byte getTTL() throws IOException {
                return 0;
            }
            
            @Override
            protected void setTimeToLive(int n) throws IOException {
            }
            
            @Override
            protected int getTimeToLive() throws IOException {
                return 0;
            }
            
            @Override
            protected void join(InetAddress inetAddress) throws IOException {
            }
            
            @Override
            protected void leave(InetAddress inetAddress) throws IOException {
            }
            
            @Override
            protected void joinGroup(SocketAddress socketAddress, NetworkInterface networkInterface) throws IOException {
            }
            
            @Override
            protected void leaveGroup(SocketAddress socketAddress, NetworkInterface networkInterface) throws IOException {
            }
            
            @Override
            protected void close() {
            }
            
            @Override
            public Object getOption(int n) throws SocketException {
                return null;
            }
            
            @Override
            public void setOption(int n, Object o) throws SocketException {
            }
        };
    }
}
